/* Copyright (c) 2006-2014 Polybus Systems Corp., Westford, MA, USA */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "sys_basher.h"
#include <sys/sysinfo.h>
#include <asm/fcntl.h>

long fileTest(char *base,long num, long num_threads,file_info *finfo,SInt64 limit,result rslt[NUM_DISK_OPS][MAX_DIRS][NUM_DISK_MODES][MAX_FILE_SIZES],SInt32 vnum)
{
  long i,j,k,l,total_errs=0,num_dirs,mode,size,num_sizes;
  char string[1024],lstring[256];
  long errors[4][32][128];
  double wbandw[4][32][128];
  double rbandw[4][32][128];
  double rlatency[4][32][128];
  double wlatency[4][32][128];
  double bandw[8],usecs;
  double length[32];
  long modes[] = {O_SYNC,0L,O_SYNC | O_DIRECT,O_DIRECT,0L};
  char *mode_name[] = {"Cached","Cached","Direct","Direct",0L};
  char *line_a[] = {"Cached Read","Cached Write","Cached Read","Cached Write","Direct Read","Direct Write","Direct Read","Direct Write",0L};
  char *line_b[] = {"Sync Mode","Sync Mode","No Sync","No Sync","Sync Mode","Sync Mode","No Sync","No Sync",0L};
  char *line_c[] = {"Sync","No Sync","Sync","No Sync",0L};
  char path[128][512];

  sprintf(string,"\nThread %ld of %ld: Running Disk Tests\n",(long)vnum,(long)num_threads);
  printTranscript(string,FALSE);

  num_dirs=0;
  while(getDirName(num,num_dirs,finfo,path[num_dirs]))
    {
      num_dirs++;
    }

  for(mode=0;mode<4;mode++)
    {
      num_sizes = 0;
      for(size=12;size<MAX_FILE_SIZES;size+=2)
	{
	  length[num_sizes] = 0x00000001LL << size;
	  if(length[num_sizes] < limit)
	    {
	      fileWrite(base,num,size,modes[mode],finfo,wbandw[mode][num_sizes],wlatency[mode][num_sizes]);
	      num_sizes++;
	    }
	}
      num_sizes = 0;
      for(size=12;size<MAX_FILE_SIZES;size+=2)
	{
	  if(length[num_sizes] < limit)
	    {
	      checkFile(base,num,size,modes[mode],mode_name[mode],finfo,rbandw[mode][num_sizes],errors[mode][num_sizes],rlatency[mode][num_sizes]);
	      num_sizes++;
	    }
	}
    }

  total_errs = 0;
  for(i=0;i<num_dirs;i++)
    {
      sprintf(string,"\nThread %ld of %ld, Directory:%s\n",(long)vnum,(long)num_threads,finfo->path[i]);
      printTranscript(string,FALSE);
      displayTitles("File Size",line_a,8L);
      displayTitles("",line_b,8L);
      for(j=0;j<num_sizes;j++)
	{
	  iscale(lstring,"B",length[j]);
	  l=0;
	  for(k=0;k<4;k++)
	    {
	      bandw[l] = rbandw[k][j][i];
	      bandw[l+1] = wbandw[k][j][i];
	      //usecs = length[j]/rbandw[k][j][i];
	      usecs = rlatency[k][j][i];
	      if(num_threads == 1) usecs *= 1000.0;
	      updateBandwidth(rbandw[k][j][i],usecs,length[j],&rslt[0][i][k][j]);
	      //	      usecs = length[j]/wbandw[k][j][i];
	      usecs = wlatency[k][j][i];
	      updateBandwidth(wbandw[k][j][i],usecs,length[j],&rslt[1][i][k][j]);
	      sprintf(rslt[0][i][k][j].name,"Read %s %s",mode_name[k],line_c[k]);
	      sprintf(rslt[1][i][k][j].name,"Write %s %s",mode_name[k],line_c[k]);
	      rslt[0][i][k][j].size = length[j];
	      rslt[1][i][k][j].size = length[j];
	      rslt[0][i][k][j].runs++;
	      rslt[1][i][k][j].runs++;
	      rslt[0][i][k][j].errors += errors[k][j][i];
	      rslt[1][i][k][j].errors += errors[k][j][i];
	      l+=2;
	    }
	  displayLine(lstring,"B/S",bandw,8L);
	}
      printTranscript("\n",FALSE);
      displayTitles("File Size",mode_name,4L);
      displayTitles("",line_c,4L);
      for(j=0;j<num_sizes;j++)
	{
	  iscale(lstring,"B",length[j]);
	  for(k=0;k<4;k++)
	    {
	      if(errors[k][j][i] == 0L)
		{
		  sprintf(string,"PASSED %s %s %s:\t%s/disk_test\n",lstring,mode_name[k],line_c[k],finfo->path[i]);
		  printTranscript(string,TRUE);
		}
	      else
		{
		  sprintf(string,"FAILED %s %s %s:\t%s/disk_test\n",lstring,mode_name[k],line_c[k],finfo->path[i]);
		  printTranscript(string,TRUE);
		  total_errs++;
		}
	    }
	}
    }
  return(total_errs == 0L);
}

void fileWrite(char *base,long num,long size,long mode,file_info *finfo,double wbandw[128],double latency[128])
{
  long fp,err,findx=0;
  char string[512],estring[512];
  SInt64 i,length;
  SInt64 *indx,value;
  double start,end,wusecs,len;
  string[0] = 0;
  while(getPathName(num,size,findx,finfo,string))
    {
      fp = open(string,O_RDWR | O_CREAT | mode , S_IRWXU | S_IRWXG | S_IRWXO);
      if(fp == -1)
	{
	  sprintf(estring,"open %s for write",string);
	  perror(estring);
	}
      else
	{
	  length = 0x00000001LL << size;
	  indx = (SInt64 *)base;
	  value = num << 16;
	  value += size;
	  
	  for(i=0;i<length;i+=8)
	    {
	      *indx++ = value;
	      value += 0x0000000100000000LL;
	    }
	  start = getTime();
	  err = write(fp,(void *)base,(size_t)length);
	  end = getTime();
	  if(err == -1)
	    {
	      sprintf(estring,"write %s",string);
	      perror(estring);
	    }
	  err = fsync(fp);
	  if(err == -1)
	    {
	      sprintf(estring,"fsync %s",string);
	      perror(estring);
	    }
	  close(fp);
	  wusecs = end - start;
	  len = length;
	  wbandw[findx] = len/wusecs;
	  latency[findx] = wusecs;
	}
      findx++;
    }
}

void checkFile(char *base,long num,long size,long mode,char *mstring,file_info *finfo,double rbandw[128],long errors[128],double latency[128])
{
  long i,length,findx=0;
  SInt64 value,exp,*indx;
  long fp,err;
  char string[512],estring[512];
  double start=0,end=0,rusecs,len;
  string[0] = 0;

  while(getPathName(num,size,findx,finfo,string))
    {
      length = 0x00000001LL << size;
      errors[findx] = 0;
      fp = open(string,O_RDONLY | mode,0L);
      if(fp == -1L)
	{
	  sprintf(estring,"open %s for read",string);
	  perror(estring);
	}
      else
	{
	  err = lseek(fp,0,SEEK_SET); 
	  if(err == -1)
	    {
	      sprintf(estring,"lseek %s",string);
	      perror(estring);
	    }
	  else
	    {
	      start = getTime();
	      end = start;
	      err = read(fp,base,length);
	      if(err == -1)
		{
		  sprintf(estring,"read %s",string);
		  perror(estring);
		}
	      else
		{

		  end = getTime();
		  indx = (SInt64 *)base;
		  exp = num << 16;
		  exp += size;
		  for(i=0;i<length;i+=8L)
		    {
		      value = *indx++;
		      if(exp != value)
			{
			  errors[findx] =  errors[findx]+1;
			}
		      exp += 0x0000000100000000LL;
		    }
		}
	    }
	  close(fp);
	  unlink(string);
	}

      rusecs = end - start;
      len = length;
      rbandw[findx] = len/rusecs;
      latency[findx] = rusecs;
      findx++;
    }

}

void createDirs(long num_threads,file_info *finfo)
{
  long num_dirs = 0,i;
  char string[1024];

  num_dirs=0;
  while(getDirName(-1L,num_dirs,finfo,string))
    {
      mkdir(string,0x01FFL);
      num_dirs++;
    }

  for(i=1;i<=num_threads;i++)
    {
      num_dirs=0;
      while(getDirName(i,num_dirs,finfo,string))
	{
	  mkdir(string,0x01FFL);
	  num_dirs++;
	}
    }
  sync();
  sync();
}

void deleteDirs(long num_threads,file_info *finfo)
{
  long num_dirs = 0,i;
  char string[1024];
  for(i=1;i<= num_threads;i++)
    {
      num_dirs=0;
      while(getDirName(i,num_dirs,finfo,string))
	{
	  rmdir(string);
	  num_dirs++;
	}
    }
  num_dirs=0;
  while(getDirName(-1L,num_dirs,finfo,string))
    {
      rmdir(string);
      num_dirs++;
    }
  sync();
  sync();
}
