/* Copyright (c) 2006-2014 Polybus Systems Corp., Westford, MA, USA */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

typedef unsigned char Boolean;

#define TRUE 1
#define FALSE 0

#define MAX_CORES 32
#define MAX_THREADS ((MAX_CORES +1) * (MAX_CORES/2)) + 1
#define MAX_DIRS 32
#define MAX_FILE_SIZES 28
#define MAX_SIZES 32

#define NUM_FIXED 16
#define NUM_BANK 16
#define NUM_ADDREQ 32
#define NUM_LFSR 8
#define NUM_WALK 8
#define NUM_REV_ADDR 2

typedef int16_t SInt16;
typedef int32_t SInt32;
typedef long long int SInt64;

typedef u_int8_t UInt8;
typedef u_int16_t UInt16;
typedef u_int32_t UInt32;
typedef long long unsigned int UInt64;


enum{
  WALK_TST0=0,
  WALK_TST1,
  WALK_TST2,
  WALK_TST3,
  WALK_TST4,
  WALK_TST5,
  WALK_TST6,
  WALK_TST7,
  FIXED_MEM0,
  FIXED_MEM1,
  FIXED_MEM2,
  FIXED_MEM3,
  FIXED_MEM4,
  FIXED_MEM5,
  FIXED_MEM6,
  FIXED_MEM7,
  FIXED_MEM8,
  FIXED_MEM9,
  FIXED_MEM10,
  FIXED_MEM11,
  FIXED_MEM12,
  FIXED_MEM13,
  FIXED_MEM14,
  FIXED_MEM15,
  ADDR_EQ0,
  ADDR_EQ1,
  ADDR_EQ2,
  ADDR_EQ3,
  ADDR_EQ4,
  ADDR_EQ5,
  ADDR_EQ6,
  ADDR_EQ7,
  ADDR_EQ8,
  ADDR_EQ9,
  ADDR_EQ10,
  ADDR_EQ11,
  ADDR_EQ12,
  ADDR_EQ13,
  ADDR_EQ14,
  ADDR_EQ15,
  ADDR_EQ16,
  ADDR_EQ17,
  ADDR_EQ18,
  ADDR_EQ19,
  ADDR_EQ20,
  ADDR_EQ21,
  ADDR_EQ22,
  ADDR_EQ23,
  ADDR_EQ24,
  ADDR_EQ25,
  ADDR_EQ26,
  ADDR_EQ27,
  ADDR_EQ28,
  ADDR_EQ29,
  ADDR_EQ30,
  ADDR_EQ31,
  BANK_TST2,
  BANK_TST3,
  BANK_TST4,
  BANK_TST5,
  BANK_TST6,
  BANK_TST7,
  BANK_TST8,
  BANK_TST9,
  BANK_TST10,
  BANK_TST11,
  BANK_TST12,
  BANK_TST13,
  BANK_TST14,
  BANK_TST15,
  LFSR_TST0,
  LFSR_TST1,
  LFSR_TST2,
  LFSR_TST3,
  LFSR_TST4,
  LFSR_TST5,
  LFSR_TST6,
  LFSR_TST7,
  REV_ADDR0,
  REV_ADDR1,
  MEM_BAND,
  INT_TST,
  FP_TST,
  DISK_TST,
  SWAP_TST,
  NUM_TESTS
};

enum{
  DISK_ONLY = 0x1,
  BAND_ONLY = 0x2,
  SWAP_ONLY = 0x4,
  FP_ONLY =      0x100L,
  INT_ONLY =     0x1000L,
  ADDR_EQ_ONLY = 0x10000L,
  BANK_ONLY =    0x20000L,
  PATTERN_ONLY = 0x40000L,
  RANDOM_ONLY =  0x80000L,
  WALKING_ONLY = 0x100000L,
  REV_ADDR_ONLY =0x200000L
};

enum{
  ADD_RR = 0,
  ADD_MR,
  MPY_RR,
  MPY_MR,
  DIV_RR,
  DIV_MR,
  DOT_MRR,
  DOT_MMR,
  NUM_FP_TESTS
};

enum{
  CACHED_SYNC=0,
  CACHED_NOSYNC,
  DIRECT_SYNC,
  DIRECT_NOSYNC,
  NUM_DISK_MODES
};

enum{
  DISK_READ=0,
  DISK_WRITE,
  NUM_DISK_OPS
};

typedef struct{
  char path[128][512];
  SInt32 num_paths;
} file_info;

typedef struct{
  SInt16 int16[MAX_SIZES][NUM_FP_TESTS];
  SInt32 int32[MAX_SIZES][NUM_FP_TESTS];
  SInt64 int64[MAX_SIZES][NUM_FP_TESTS];
  float fp[MAX_SIZES][NUM_FP_TESTS];
  double dbl[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale16[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale32[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale64[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale_fp[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale_dbl[MAX_SIZES][NUM_FP_TESTS];
} exp_data;

typedef struct{
  SInt64 runs;
  SInt64 errors;
  SInt64 size;
  double min;
  double max;
  double total;
  double total_time;
  char name[64];
} result;

typedef struct{
  result fixed[NUM_FIXED];
  result addreq[NUM_ADDREQ];
  result bank[NUM_BANK];
  result lfsr[NUM_LFSR];
  result walk[NUM_WALK];
  result rev_addr[NUM_REV_ADDR];
  result mem_bandwidth[2][MAX_SIZES];
  result swap[1];
  result int16_rslt[NUM_FP_TESTS][MAX_SIZES];
  result int32_rslt[NUM_FP_TESTS][MAX_SIZES];
  result int64_rslt[NUM_FP_TESTS][MAX_SIZES];
  result sgl_rslt[NUM_FP_TESTS][MAX_SIZES];
  result dbl_rslt[NUM_FP_TESTS][MAX_SIZES];
  result disk_rslt[NUM_DISK_OPS][MAX_DIRS][NUM_DISK_MODES][MAX_FILE_SIZES];
} result_summary;

typedef struct{
  Boolean rev_addr;
  Boolean lfsr;
  Boolean walk;
  Boolean addreq;
  Boolean bank;
  Boolean dbl;
  Boolean disk;
  Boolean fixed;
  Boolean int16;
  Boolean int32;
  Boolean int64;
  Boolean mem_bw;
  Boolean sgl;
  Boolean swap;
} run_summary;


typedef struct{
  SInt64 limit;
  SInt64 long_limit;
  UInt64 *seed;
  void *base;
  SInt32 errors;
  SInt32 num;
  SInt32 vnum;
  SInt32 num_threads;
  SInt32 test;
  SInt32 firsttime;
  SInt32 verbose;
  exp_data *expected;
  file_info *finfo;
  result_summary *rslt_summary;
} thread_data;

typedef struct{
  SInt64 totalram;
  SInt64 freeram;
  SInt64 totalswap;
  SInt64 freeswap;
  SInt64 bufferram;
  SInt64 sharedram;
  SInt64 mem_unit;
} my_sysinfo;

typedef struct{
  time_t sec;
  suseconds_t usecs;
} time_def;

typedef struct {
  int tm_sec;         /* seconds */
  int tm_min;         /* minutes */
  int tm_hour;        /* hours */
  int tm_mday;        /* day of the month */
  int tm_mon;         /* month */
  int tm_year;        /* year */
  int tm_wday;        /* day of the week */
  int tm_yday;        /* day in the year */
  int tm_isdst;       /* daylight saving time */
} time_of_day;

typedef struct {
  UInt64 *select;
  double *minutes;
  SInt64 *dutycycle;
  file_info *finfo;
  Boolean *use_total;
  long *threads;
} config_args;

typedef struct
{
  UInt8 type;
  UInt8 length;
  UInt16 handle;
  UInt16 dev_handle;
  UInt16 mem_handle;
  UInt16 row_pos;
  UInt16 interleave_pos;
  UInt16 interleave_depth;
  UInt16 partition_width;
  UInt32 num_strings;
  UInt64 start_addr;
  UInt64 end_addr;
  UInt16 total_width;
  UInt16 data_width;
  UInt8 device_set;
  UInt16 type_detail;
  UInt16 speed;
  UInt8 attributes;
  UInt64 size;
  UInt16 configured_speed;
  char *asset_tag;
  char *bank_locator;
  char *boot_status;
  char *device_locator;
  char *form_factor;
  char *manufacturer;
  char *mem_type;
  char *part_number;
  char *serial_num;
  char mem_detail[1024];
  char *data[256];
} dmi_entry;

enum{
  DINFO_TYPE = 0x0,
  DINFO_LENGTH = 0x1,
  DINFO_HANDLE = 0x2,
  DINFO_MEM_HANDLE = 0x4,
  DINFO_ERR_HANDLE = 0x6,
  DINFO_TOTAL_WIDTH = 0x8,
  DINFO_DATA_WIDTH = 0xA,
  DINFO_SIZE = 0xC,
  DINFO_FORM_FACTOR = 0xE,
  DINFO_DEVICE_SET = 0xF,
  DINFO_DEVICE_LOCATOR = 0x10,
  DINFO_BANK_LOCATOR = 0x11,
  DINFO_MEM_TYPE = 0x12,
  DINFO_TYPE_DETAIL = 0x13,
  DINFO_SPEED = 0x15,
  DINFO_MANUFACTURER = 0x17,
  DINFO_SERIAL_NUM = 0x18,
  DINFO_ASSET_TAG = 0x19,
  DINFO_PART_NUMBER = 0x1A,
  DINFO_ATTRIBUTES = 0x1B,
  DINFO_EXT_SIZE = 0x1C,
  DINFO_CONFIGURED_SPEED = 0x20
};

enum{
  DINFO_LEN_TYPE = 1,
  DINFO_LEN_LENGTH = 1,
  DINFO_LEN_HANDLE = 2,
  DINFO_LEN_MEM_HANDLE = 2,
  DINFO_LEN_ERR_HANDLE = 2,
  DINFO_LEN_TOTAL_WIDTH = 2,
  DINFO_LEN_DATA_WIDTH = 2,
  DINFO_LEN_SIZE = 2,
  DINFO_LEN_FORM_FACTOR = 1,
  DINFO_LEN_DEVICE_SET = 1,
  DINFO_LEN_DEVICE_LOCATOR = 1,
  DINFO_LEN_BANK_LOCATOR = 1,
  DINFO_LEN_MEM_TYPE = 1,
  DINFO_LEN_TYPE_DETAIL = 2,
  DINFO_LEN_SPEED = 2,
  DINFO_LEN_MANUFACTURER = 1,
  DINFO_LEN_SERIAL_NUM = 1,
  DINFO_LEN_ASSET_TAG = 1,
  DINFO_LEN_PART_NUMBER = 1,
  DINFO_LEN_ATTRIBUTES = 1,
  DINFO_LEN_EXT_SIZE = 4,
  DINFO_LEN_CONFIGURED_SPEED = 2
};

enum{
  MEM_DEVICE_TYPE = 0x0,
  MEM_DEVICE_LENGTH = 0x1,
  MEM_DEVICE_HANDLE = 0x2,
  MEM_DEVICE_START_ADDR = 0x4,
  MEM_DEVICE_END_ADDR = 0x8,
  MEM_DEVICE_DEV_HANDLE = 0xC,
  MEM_DEVICE_MEM_HANDLE = 0xE,
  MEM_DEVICE_ROW_POS = 0x10,
  MEM_DEVICE_INTERLEAVE_POS = 0x11,
  MEM_DEVICE_INTERLEAVE_DEPTH = 0x12,
  MEM_DEVICE_EXT_START_ADDR = 0x13,
  MEM_DEVICE_EXT_END_ADDR = 0x1B
};  

enum{
  MEM_DEVICE_LEN_TYPE = 1,
  MEM_DEVICE_LEN_LENGTH = 1,
  MEM_DEVICE_LEN_HANDLE = 2,
  MEM_DEVICE_LEN_START_ADDR = 4,
  MEM_DEVICE_LEN_END_ADDR = 4,
  MEM_DEVICE_LEN_DEV_HANDLE = 2,
  MEM_DEVICE_LEN_MEM_HANDLE = 2,
  MEM_DEVICE_LEN_ROW_POS = 1,
  MEM_DEVICE_LEN_INTERLEAVE_POS = 1,
  MEM_DEVICE_LEN_INTERLEAVE_DEPTH = 1,
  MEM_DEVICE_LEN_EXT_START_ADDR = 8,
  MEM_DEVICE_LEN_EXT_END_ADDR = 8
};  

enum{
  MEM_ARRAY_TYPE = 0x0,
  MEM_ARRAY_LENGTH = 0x1,
  MEM_ARRAY_HANDLE = 0x2,
  MEM_ARRAY_START_ADDR = 0x4,
  MEM_ARRAY_END_ADDR = 0x8,
  MEM_ARRAY_MEM_HANDLE = 0xC,
  MEM_ARRAY_PARTITION_WIDTH = 0xE,
  MEM_ARRAY_EXT_START_ADDR = 0xF,
  MEM_ARRAY_EXT_END_ADDR = 0x17
};

enum{
  MEM_ARRAY_LEN_TYPE = 1,
  MEM_ARRAY_LEN_LENGTH = 1,
  MEM_ARRAY_LEN_HANDLE = 2,
  MEM_ARRAY_LEN_START_ADDR = 4,
  MEM_ARRAY_LEN_END_ADDR = 4,
  MEM_ARRAY_LEN_MEM_HANDLE = 2,
  MEM_ARRAY_LEN_PARTITION_WIDTH = 1,
  MEM_ARRAY_LEN_EXT_START_ADDR = 8,
  MEM_ARRAY_LEN_EXT_END_ADDR = 8
};
  
enum{
  COMMON_TYPE = 0,
  COMMON_LENGTH = 1,
  COMMON_HANDLE = 2
};

enum{
  COMMON_LEN_TYPE = 1,
  COMMON_LEN_LENGTH = 1,
  COMMON_LEN_HANDLE = 2
};

enum{
  SYS_BOOT_TYPE = 0,
  SYS_BOOT_LENGTH = 1,
  SYS_BOOT_HANDLE = 2,
  SYS_BOOT_BOOT_STATUS = 0xA,
};

enum{
  SYS_BOOT_LEN_TYPE = 1,
  SYS_BOOT_LEN_LENGTH = 1,
  SYS_BOOT_LEN_HANDLE = 2,
  SYS_BOOT_LEN_BOOT_STATUS = 1
};


enum{
  CPU_INFO = 4,
  DIMM_INFORMATION = 17,
  MEM_ARRAY = 19,
  MEM_DEVICE = 20,
  SYS_BOOT_STAT = 32
};

enum{
  NUM_FORM_FACTORS = 17,
  NUM_MEM_TYPES = 0x1A,
  NUM_BOOT_STAT = 9
};

 

//Page map entry
// Bits 0-54  page frame number (PFN) if present
// Bits 0-4   swap type if swapped
// Bits 5-54  swap offset if swapped
// Bit  55    pte is soft-dirty (see Documentation/vm/soft-dirty.txt)
// Bits 56-60 zero
// Bit  61    page is file-page or shared-anon
// Bit  62    page swapped
// Bit  63    page present
enum{
  PFN = 0x7FFFFFFFFFFFFFLL,
  SWAP_OFF_MASK = 0x1FLL,
  SWAP_OFF_SHF = 5L,
  SOFT_DIRTY_BIT = 55L,
  FILE_PAGE_BIT = 61L,
  SWAP_BIT = 62L,
  PRESENT_BIT = 63L,
  SWAP_MASK = 0x4000000000000000LL,
  PRESENT_MASK = 0x8000000000000000LL
};
